/*

Mat 259 Project 2 3D Visualization: "Happiness Maps of the Seattle Public Library"
9 March 2023
Processing 4.0.1
                                                     
Author: Jenni Hutson                  
Supervisor: George Legrady                            

Purpose: To show how happiness trends can be determined and mapped from the titles checked out each day at the SPL.     

Usage: 1. A mouse left-drag will rotate the camera around the subject.
          2. A right drag will zoom in and out. 
          3. A middle-drag (command-left-drag on mac) will pan. 
          4. A double-click restores the camera to its original position. 
          5. With SelectYear turned on, use up and down arrows to cycle through layers
          6. With SelectYear turned on, use + and - keys to increase or decrease the vertical scaling of selected layer.

*/

import peasy.*;
import controlP5.*;

PeasyCam cam;
ControlP5 cp5;

ArrayList<Year> years = new ArrayList<Year>();
ArrayList<Month> months = new ArrayList<Month>();

String[] monthNames = {"January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"};

Table table;
int numRows;

int startYear = 2005;
int startMonth = -1;

int startSelect = 0;

float xInc = 8;
float yInc = 20;
float zInc = 20;

//int xOriginal = -130;
int xOriginal = (displayWidth/2)-(((int)xInc*28)/2);
int yOriginal = 0;
int yOg = yOriginal;
int zOriginal = -100;

Float valley = 0.0;
Float peak = 600.0;
Float heightMult = 1.0;

int lowestBooks = 100000;
int highestBooks = -100;
Float lowest = 1000.0;
Float highest = -1000.0;

int count_threshold = 100;

Float avgHappiness = 0.0;
int avgCount = 0;

int YEAR_COUNT = 0;

boolean showBox = true;
boolean showLabels = true;
boolean selectYear = false;
boolean onlySelectedYear = false;

void setup() {
  colorMode(HSB, 100);
  size(displayWidth, displayHeight, P3D);
  
  cp5 = new ControlP5(this);
  
  cam = new PeasyCam(this, 550);
  cam.setMaximumDistance(700);
  cam.rotateY(-0.4); 
  
  PFont font = createFont("Arial", 20);
  textFont(font);
  textSize(14);
  
  table = loadTable("HappinessPerDay2006_2022.csv", "header");
  numRows = table.getRowCount();
  println("Rows: " + numRows);
  
  int yearI = -1;
  int currYear = startYear;
  int currMonth = startMonth;
  int lastDay = -1;

  for (int i=0; i<numRows; i++){
    String date = table.getString(i,0);
    int year = Integer.parseInt(date.substring(0,4));
    Float happiness = table.getFloat(i,1);
    int book_count = table.getInt(i,2);
    int month = Integer.parseInt(date.substring(5,7))-1;
    int day = Integer.parseInt(date.substring(8,10))-1;
    
    if (book_count < lowestBooks) {
      lowestBooks = book_count;
    } 
    if (book_count > highestBooks) {
      highestBooks = book_count;
    }
    
    if (happiness < lowest && book_count > count_threshold) {
      lowest = happiness;
    }
    if (happiness > highest) {
      highest = happiness;
    }
    
    if (year != currYear) {
      years.add(new Year(year, new ArrayList<Month>()));
      yearI++;
      currYear = year;
    } 
    
    if(month != currMonth) {
      //if new year
      if (month < currMonth) {
        currMonth = 0;
      } else {
      //if same year
        currMonth++;
      }
      
      if (currMonth%2 == 0) {
        years.get(yearI).months.add(new Month(i, new ArrayList<Day>(),31));
      } else {
        if (currMonth == 1) {
          if (currYear%4 == 0) {
             years.get(yearI).months.add(new Month(i, new ArrayList<Day>(), 29));
          }
          else {
            years.get(yearI).months.add(new Month(i, new ArrayList<Day>(), 28));
          }
      } else if (currMonth == 7) {
        years.get(yearI).months.add(new Month(i, new ArrayList<Day>(),31));
      } else {
        years.get(yearI).months.add(new Month(i, new ArrayList<Day>(), 30));
        }
      }
    }
    
    if (day != lastDay+1) {
      //if not in same month
      if (day < lastDay+1) { //flawed logic could only have 1 datapoint in a month
        int lastMonth = currMonth-1;
        int lastYear = yearI;
        if (lastMonth <0){
          lastMonth = 11;
          lastYear = yearI-1;
        }
        int numDays = years.get(lastYear).months.get(lastMonth).num_days;
        if (lastDay < numDays-1) {
          int lasti = lastDay;
          for (int k =0; k<(numDays-(lasti+1)); k++) {
            String monthStr = String.valueOf(currMonth);
            if (currMonth+1 < 10) {
              monthStr = "0" + String.valueOf(currMonth);
            }
            String dayStr = String.valueOf(lasti+k+2);
            if ((lasti + k + 2) < 10) {
              dayStr = "0" + String.valueOf(lasti+k+2);
            }
            String newDate = String.valueOf(currYear) + "-" +monthStr + "-" +dayStr;
            Day dayI = new Day(newDate, 0.0, 0);
            years.get(lastYear).months.get(lastMonth).days.add(dayI);
            lastDay++;
          }
        }
        lastDay = -1;
      }else {
        //if still in same month 
        int lasty = lastDay;
        for (int j=0; j<(day-(lasty+1));j++){
          String monthStr = String.valueOf(currMonth+1);
          if (currMonth+1 < 10) {
            monthStr = "0" + String.valueOf(currMonth+1);
          }
          String dayStr = String.valueOf(lasty+j+2);
          if ((lasty + j + 2) < 10) { // should be <9?
            
            dayStr = "0" + String.valueOf(lasty+j+2);
          }
          String newDate = String.valueOf(currYear) + "-" +monthStr + "-" +dayStr;
          Day dayI = new Day(newDate, 0.0, 0);
          years.get(yearI).months.get(currMonth).days.add(dayI);
          lastDay++;
        } 
      }
    }
    Day dayI = new Day(date, happiness, book_count);
    
    if (book_count > count_threshold) {  
      years.get(yearI).months.get(currMonth).days.add(dayI); 
      avgHappiness += happiness;
      avgCount++;
      
    } else {
      years.get(yearI).months.get(currMonth).days.add(new Day(date, 0.0, 0));
    }
    lastDay++;
  }
  
  //if final month data ends too early
  int numDays = years.get(yearI).months.get(currMonth).num_days;
  if (lastDay < numDays) {
    for (int g =0; g<(numDays-(lastDay+1)); g++) {
      String monthStr = String.valueOf(currMonth+1);
      if (currMonth+1 < 10) {
        monthStr = "0" + String.valueOf(currMonth+1);
      }
      String dayStr = String.valueOf(lastDay+g+2);
      if ((lastDay + g + 1) < 10) {
        dayStr = "0" + String.valueOf(lastDay+g+2);
      }
      String newDate = String.valueOf(currYear) + "-" +monthStr + "-" +dayStr;
      Day dayI = new Day(newDate, 0.0, 0);
      years.get(yearI).months.get(currMonth).days.add(dayI);
    }
  }
  avgHappiness = avgHappiness/avgCount;
  YEAR_COUNT = years.size();
  
  yOriginal = -(displayHeight/35) - 35;
  println("yOriginal: ", yOriginal);
  yOg = yOriginal;

  
  ArrayList<Year> newYears = new ArrayList<Year>();
  for (int i=0; i<years.size(); i++){
    newYears.add(years.get(years.size()-i-1));
  }
  years = newYears;
  
  setGui();
}

void draw() {
  //background(240);
  background(0,0,94);
  //fill(180, 180, 180, 50);
  
  yOriginal = yOg;
  drawAllYears(); 

  yOriginal = yOg;
  drawMouseLabels();
  
  if (showBox) {
    drawBox();
  }
  if (showLabels) {
    drawLabels();
  }
  
  gui();
  if((mouseX<180) & (mouseY<180)) {
    cam.setActive(false);
  } else {
    cam.setActive(true);
  }
}

void drawAllYears() {
  float avgHeight = map(avgHappiness, lowest, highest, valley, peak);
  //every year
  for (int yearI=0; yearI<years.size(); yearI++){ //y original will change
  //years do go backwards but thats fine? 
    int val = (int)(((float)yearI/YEAR_COUNT)*100);
    if (selectYear) {
      if (startSelect ==yearI) {
        fill(val, 20, 80, 40);
      } else {
        if (onlySelectedYear) {
          yOriginal+=yInc;
          continue;
        }
        fill(0, 0, 80, 15);
      }
    } else {
      fill(val, 20, 80, 30);
    }
    beginShape(TRIANGLE_STRIP);
    //every month
    for (int monthI=0; monthI<years.get(yearI).months.size()-1; monthI++) { //will affect z
      //every day
      //for (int dayI=0; dayI<years.get(yearI).months.get(monthI).days.size(); dayI++){
      for (int dayI=0; dayI<28; dayI++){ //will affect x
       Day currDay = years.get(yearI).months.get(monthI).days.get(dayI);
       Day nextDay = years.get(yearI).months.get(monthI+1).days.get(dayI);
       float dayHeight = map(currDay.happiness, lowest, highest, valley, peak);
       if (selectYear && startSelect ==yearI) {
          dayHeight = map(currDay.happiness, lowest, highest, valley, peak*heightMult);
       }
       if (currDay.happiness <= 0.0) { //want like the average
         if (selectYear && startSelect ==yearI) {
           dayHeight = map(avgHappiness, lowest, highest, valley, peak*heightMult);
         } else {
           dayHeight = map(avgHappiness, lowest, highest, valley, peak);
         }
       } 
       float nextDayHeight = map(nextDay.happiness, lowest, highest, valley, peak);
       if (selectYear && startSelect ==yearI) {
         nextDayHeight = map(nextDay.happiness, lowest, highest, valley, peak*heightMult);
       }
       if (nextDay.happiness <= 0.0) {
         if (selectYear && startSelect ==yearI) {
           nextDayHeight = map(avgHappiness, lowest, highest, valley, peak*heightMult);
         } else {
           nextDayHeight = map(avgHappiness, lowest, highest, valley, peak);
         }
       }
       
       float bookScale = map(currDay.book_count, lowestBooks, highestBooks, 0, 100);
       float nextBookScale = map(nextDay.book_count, lowestBooks, highestBooks, 0, 100);
       
       strokeWeight(1.5);
       if (selectYear) {
          if (startSelect ==yearI) {
            stroke(bookScale, 20, 80);
          } else {
            bookScale = map(currDay.book_count, lowestBooks, highestBooks, 15, 70);
            stroke(0,0,100-bookScale);
          }
       } else {
         stroke(bookScale, 20, 80);
       }

       if (selectYear && startSelect ==yearI) {
         vertex((dayI * xInc) + xOriginal, yOriginal - dayHeight+(avgHeight*(heightMult-1)), (monthI * zInc) + zOriginal);
       } else {
         vertex((dayI * xInc) + xOriginal, yOriginal - dayHeight, (monthI * zInc) + zOriginal);
       }


      if (selectYear) {
        if (startSelect ==yearI) {
          stroke(bookScale, 20, 80);
        } else {
          nextBookScale = map(nextDay.book_count, lowestBooks, highestBooks, 15, 70);
          stroke(0,0,100-nextBookScale);
        }
       } else {
         stroke(nextBookScale, 20, 80);
       }
       //stroke(100-nextBookScale, 20, 80);
       if (selectYear && startSelect ==yearI) {
         vertex((dayI * xInc) + xOriginal, yOriginal - nextDayHeight +(avgHeight*(heightMult-1)), ((monthI+1) * zInc) + zOriginal); 
       } else {
         vertex((dayI * xInc) + xOriginal, yOriginal - nextDayHeight, ((monthI+1) * zInc) + zOriginal); 
       }
       

      }
    }
    endShape();
    yOriginal+=yInc;
  }
}

void drawMouseLabels() {
  float avgHeight = map(avgHappiness, lowest, highest, valley, peak);
  for (int yearI=0; yearI<years.size(); yearI++){ 
    if (selectYear && startSelect != yearI) {
      if (onlySelectedYear) {
          yOriginal+=yInc;
          continue;
        }
    }
    for (int monthI=0; monthI<years.get(yearI).months.size()-1; monthI++) {
      for (int dayI=0; dayI<28; dayI++){
        Day currDay = years.get(yearI).months.get(monthI).days.get(dayI);
        float dayHeight = map(currDay.happiness, lowest, highest, valley, peak);
        if (selectYear && startSelect ==yearI) {
          dayHeight = map(currDay.happiness, lowest, highest, valley, peak*heightMult);
         }
        
        if (currDay.happiness <= 0.0) { //want like the average
          if (selectYear && startSelect ==yearI) {
             dayHeight = map(avgHappiness, lowest, highest, valley, peak*heightMult);
           } else {
             dayHeight = map(avgHappiness, lowest, highest, valley, peak);
           }
        } 
        
        float y;
        if (selectYear && startSelect ==yearI) {
           y = yOriginal - dayHeight+(avgHeight*(heightMult-1));
        } else {
           y = yOriginal - dayHeight;
        }
         
        float x = (dayI * xInc) + xOriginal;
        float z = (monthI * zInc) + zOriginal; 
        
        float mouseObjectDistance = sq(mouseX-screenX(x,y,z)) + sq(mouseY-screenY(x,y,z));
        if (mouseObjectDistance < 35) {
         fill(40);
         textSize(18);
         String label = currDay.date + " ";
         label += "Happiness rating: " + currDay.happiness;
         
        hint(DISABLE_DEPTH_TEST);
        cam.beginHUD();
        text(label, mouseX, mouseY);
        cam.endHUD();
        hint(ENABLE_DEPTH_TEST);
        }
      }
    }
    yOriginal+=yInc;
  }
}

void drawBox() {
  fill(0,0,55,5);
  noStroke();
  pushMatrix();
  translate(0, yOg+30, 10);
  box(10+xInc*28,30+yInc*YEAR_COUNT, 20+zInc*12);
  popMatrix(); 
}

void drawLabels(){
  yOriginal = yOg;
  
  fill(40);
  textSize(7);
  float avgHeight = map(avgHappiness, lowest, highest, valley, peak);
  
  //years
  for (int yearI=0; yearI<years.size(); yearI++){
    String yr = String.valueOf(years.get(yearI).year);
    text(yr, xOriginal-(displayWidth/35), yOriginal + yInc*yearI - avgHeight, zOriginal);
  }
  
  //months
  for (int monthI=0;monthI<12; monthI++) {
    String mon = monthNames[monthI];
    text(mon, xOriginal-(displayWidth/35), yOriginal+(yInc*(YEAR_COUNT-1))-(avgHeight*.75), zOriginal + zInc*monthI);
  }
  
  //days
  
}

void setGui(){
  PFont pfont= createFont("Arial", 16, true);
  ControlFont font = new ControlFont(pfont, 12);
  
  float X = 60;
  float distance = 40;
  int index = 1;
  
  cp5.addToggle("showBox")
  .setPosition(X, distance*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,255,0));
  index++;
  
  cp5.addToggle("showLabels")
  .setPosition(X, distance*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,255,0));
  index++;
  
  cp5.addToggle("selectYear")
  .setPosition(X, distance*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,255,0));
  index++;
  
  cp5.addToggle("onlySelectedYear")
  .setPosition(X, distance*index)
  .setFont(font)
  .setColorCaptionLabel(color(0,255,0));
  index++;
  
  cp5.addTextlabel("label")
  .setText("With SelectYear turned on, \nuse up and down arrows to \ncycle through layers.")
  .setPosition(X-5, 5+distance*index)
  .setFont(new ControlFont(pfont, 14))
  .setColorValue(color(0,0,20));
  index++;
  
  cp5.addTextlabel("label1")
  .setText("With SelectYear turned on, \nuse + and - keys to increase \nor decrease the vertical \nscaling of selected layer.")
  .setPosition(X-5, 20+distance*index)
  .setFont(new ControlFont(pfont, 14))
  .setColorValue(color(0,0,20));
  
  cp5.setAutoDraw(false);
}

void gui() {
  hint(DISABLE_DEPTH_TEST);
  cam.beginHUD();
  cp5.draw();
  cam.endHUD();
  hint(ENABLE_DEPTH_TEST);
}
